# help.py - help utilities for PythonWin.
import win32api
import win32con
import win32ui
import string
import sys
import regutil
import string, os

htmlhelp_handle = None

html_help_command_translators = {
    win32con.HELP_CONTENTS: 1,  # HH_DISPLAY_TOC
    win32con.HELP_CONTEXT: 15,  # HH_HELP_CONTEXT
    win32con.HELP_FINDER: 1,  # HH_DISPLAY_TOC
}


def FinalizeHelp():
    global htmlhelp_handle
    if htmlhelp_handle is not None:
        import win32help

        try:
            # frame = win32ui.GetMainFrame().GetSafeHwnd()
            frame = 0
            win32help.HtmlHelp(frame, None, win32help.HH_UNINITIALIZE, htmlhelp_handle)
        except win32help.error:
            print("Failed to finalize htmlhelp!")
        htmlhelp_handle = None


def OpenHelpFile(fileName, helpCmd=None, helpArg=None):
    "Open a help file, given a full path"
    # default help arg.
    win32ui.DoWaitCursor(1)
    try:
        if helpCmd is None:
            helpCmd = win32con.HELP_CONTENTS
        ext = os.path.splitext(fileName)[1].lower()
        if ext == ".hlp":
            win32api.WinHelp(
                win32ui.GetMainFrame().GetSafeHwnd(), fileName, helpCmd, helpArg
            )
        # XXX - using the htmlhelp API wreaks havoc with keyboard shortcuts
        # so we disable it, forcing ShellExecute, which works fine (but
        # doesn't close the help file when Pythonwin is closed.
        # Tom Heller also points out http://www.microsoft.com/mind/0499/faq/faq0499.asp,
        # which may or may not be related.
        elif 0 and ext == ".chm":
            import win32help

            global htmlhelp_handle
            helpCmd = html_help_command_translators.get(helpCmd, helpCmd)
            # frame = win32ui.GetMainFrame().GetSafeHwnd()
            frame = 0  # Dont want it overlapping ours!
            if htmlhelp_handle is None:
                htmlhelp_hwnd, htmlhelp_handle = win32help.HtmlHelp(
                    frame, None, win32help.HH_INITIALIZE
                )
            win32help.HtmlHelp(frame, fileName, helpCmd, helpArg)
        else:
            # Hope that the extension is registered, and we know what to do!
            win32api.ShellExecute(0, "open", fileName, None, "", win32con.SW_SHOW)
        return fileName
    finally:
        win32ui.DoWaitCursor(-1)


def ListAllHelpFiles():
    ret = []
    ret = _ListAllHelpFilesInRoot(win32con.HKEY_LOCAL_MACHINE)
    # Ensure we don't get dups.
    for item in _ListAllHelpFilesInRoot(win32con.HKEY_CURRENT_USER):
        if item not in ret:
            ret.append(item)
    return ret


def _ListAllHelpFilesInRoot(root):
    """Returns a list of (helpDesc, helpFname) for all registered help files"""
    import regutil

    retList = []
    try:
        key = win32api.RegOpenKey(
            root, regutil.BuildDefaultPythonKey() + "\\Help", 0, win32con.KEY_READ
        )
    except win32api.error as exc:
        import winerror

        if exc.winerror != winerror.ERROR_FILE_NOT_FOUND:
            raise
        return retList
    try:
        keyNo = 0
        while 1:
            try:
                helpDesc = win32api.RegEnumKey(key, keyNo)
                helpFile = win32api.RegQueryValue(key, helpDesc)
                retList.append((helpDesc, helpFile))
                keyNo = keyNo + 1
            except win32api.error as exc:
                import winerror

                if exc.winerror != winerror.ERROR_NO_MORE_ITEMS:
                    raise
                break
    finally:
        win32api.RegCloseKey(key)
    return retList


def SelectAndRunHelpFile():
    from pywin.dialogs import list

    helpFiles = ListAllHelpFiles()
    if len(helpFiles) == 1:
        # only 1 help file registered - probably ours - no point asking
        index = 0
    else:
        index = list.SelectFromLists("Select Help file", helpFiles, ["Title"])
    if index is not None:
        OpenHelpFile(helpFiles[index][1])


helpIDMap = None


def SetHelpMenuOtherHelp(mainMenu):
    """Modifies the main Help Menu to handle all registered help files.
    mainMenu -- The main menu to modify - usually from docTemplate.GetSharedMenu()
    """

    # Load all help files from the registry.
    global helpIDMap
    if helpIDMap is None:
        helpIDMap = {}
        cmdID = win32ui.ID_HELP_OTHER
        excludeList = ["Main Python Documentation", "Pythonwin Reference"]
        firstList = ListAllHelpFiles()
        # We actually want to not only exclude these entries, but
        # their help file names (as many entries may share the same name)
        excludeFnames = []
        for desc, fname in firstList:
            if desc in excludeList:
                excludeFnames.append(fname)

        helpDescs = []
        for desc, fname in firstList:
            if fname not in excludeFnames:
                helpIDMap[cmdID] = (desc, fname)
                win32ui.GetMainFrame().HookCommand(HandleHelpOtherCommand, cmdID)
                cmdID = cmdID + 1

    helpMenu = mainMenu.GetSubMenu(
        mainMenu.GetMenuItemCount() - 1
    )  # Help menu always last.
    otherHelpMenuPos = 2  # cant search for ID, as sub-menu has no ID.
    otherMenu = helpMenu.GetSubMenu(otherHelpMenuPos)
    while otherMenu.GetMenuItemCount():
        otherMenu.DeleteMenu(0, win32con.MF_BYPOSITION)

    if helpIDMap:
        for id, (desc, fname) in helpIDMap.items():
            otherMenu.AppendMenu(win32con.MF_ENABLED | win32con.MF_STRING, id, desc)
    else:
        helpMenu.EnableMenuItem(
            otherHelpMenuPos, win32con.MF_BYPOSITION | win32con.MF_GRAYED
        )


def HandleHelpOtherCommand(cmd, code):
    OpenHelpFile(helpIDMap[cmd][1])



import sys
import string, tokenize
from . import PyParse
from pywin import default_scintilla_encoding

if sys.version_info < (3,):
    # in py2k, tokenize() takes a 'token eater' callback, while
    # generate_tokens is a generator that works with str objects.
    token_generator = tokenize.generate_tokens
else:
    # in py3k tokenize() is the generator working with 'byte' objects, and
    # token_generator is the 'undocumented b/w compat' function that
    # theoretically works with str objects - but actually seems to fail)
    token_generator = tokenize.tokenize


class AutoIndent:

    menudefs = [
        (
            "edit",
            [
                None,
                ("_Indent region", "<<indent-region>>"),
                ("_Dedent region", "<<dedent-region>>"),
                ("Comment _out region", "<<comment-region>>"),
                ("U_ncomment region", "<<uncomment-region>>"),
                ("Tabify region", "<<tabify-region>>"),
                ("Untabify region", "<<untabify-region>>"),
                ("Toggle tabs", "<<toggle-tabs>>"),
                ("New indent width", "<<change-indentwidth>>"),
            ],
        ),
    ]

    keydefs = {
        "<<smart-backspace>>": ["<Key-BackSpace>"],
        "<<newline-and-indent>>": ["<Key-Return>", "<KP_Enter>"],
        "<<smart-indent>>": ["<Key-Tab>"],
    }

    windows_keydefs = {
        "<<indent-region>>": ["<Control-bracketright>"],
        "<<dedent-region>>": ["<Control-bracketleft>"],
        "<<comment-region>>": ["<Alt-Key-3>"],
        "<<uncomment-region>>": ["<Alt-Key-4>"],
        "<<tabify-region>>": ["<Alt-Key-5>"],
        "<<untabify-region>>": ["<Alt-Key-6>"],
        "<<toggle-tabs>>": ["<Alt-Key-t>"],
        "<<change-indentwidth>>": ["<Alt-Key-u>"],
    }

    unix_keydefs = {
        "<<indent-region>>": [
            "<Alt-bracketright>",
            "<Meta-bracketright>",
            "<Control-bracketright>",
        ],
        "<<dedent-region>>": [
            "<Alt-bracketleft>",
            "<Meta-bracketleft>",
            "<Control-bracketleft>",
        ],
        "<<comment-region>>": ["<Alt-Key-3>", "<Meta-Key-3>"],
        "<<uncomment-region>>": ["<Alt-Key-4>", "<Meta-Key-4>"],
        "<<tabify-region>>": ["<Alt-Key-5>", "<Meta-Key-5>"],
        "<<untabify-region>>": ["<Alt-Key-6>", "<Meta-Key-6>"],
        "<<toggle-tabs>>": ["<Alt-Key-t>"],
        "<<change-indentwidth>>": ["<Alt-Key-u>"],
    }

    # usetabs true  -> literal tab characters are used by indent and
    #                  dedent cmds, possibly mixed with spaces if
    #                  indentwidth is not a multiple of tabwidth
    #         false -> tab characters are converted to spaces by indent
    #                  and dedent cmds, and ditto TAB keystrokes
    # indentwidth is the number of characters per logical indent level.
    # tabwidth is the display width of a literal tab character.
    # CAUTION:  telling Tk to use anything other than its default
    # tab setting causes it to use an entirely different tabbing algorithm,
    # treating tab stops as fixed distances from the left margin.
    # Nobody expects this, so for now tabwidth should never be changed.
    usetabs = 1
    indentwidth = 4
    tabwidth = 8  # for IDLE use, must remain 8 until Tk is fixed

    # If context_use_ps1 is true, parsing searches back for a ps1 line;
    # else searches for a popular (if, def, ...) Python stmt.
    context_use_ps1 = 0

    # When searching backwards for a reliable place to begin parsing,
    # first start num_context_lines[0] lines back, then
    # num_context_lines[1] lines back if that didn't work, and so on.
    # The last value should be huge (larger than the # of lines in a
    # conceivable file).
    # Making the initial values larger slows things down more often.
    num_context_lines = 50, 500, 5000000

    def __init__(self, editwin):
        self.editwin = editwin
        self.text = editwin.text

    def config(self, **options):
        for key, value in options.items():
            if key == "usetabs":
                self.usetabs = value
            elif key == "indentwidth":
                self.indentwidth = value
            elif key == "tabwidth":
                self.tabwidth = value
            elif key == "context_use_ps1":
                self.context_use_ps1 = value
            else:
                raise KeyError("bad option name: %s" % repr(key))

    # If ispythonsource and guess are true, guess a good value for
    # indentwidth based on file content (if possible), and if
    # indentwidth != tabwidth set usetabs false.
    # In any case, adjust the Text widget's view of what a tab
    # character means.

    def set_indentation_params(self, ispythonsource, guess=1):
        if guess and ispythonsource:
            i = self.guess_indent()
            if 2 <= i <= 8:
                self.indentwidth = i
            if self.indentwidth != self.tabwidth:
                self.usetabs = 0

        self.editwin.set_tabwidth(self.tabwidth)

    def smart_backspace_event(self, event):
        text = self.text
        first, last = self.editwin.get_selection_indices()
        if first and last:
            text.delete(first, last)
            text.mark_set("insert", first)
            return "break"
        # Delete whitespace left, until hitting a real char or closest
        # preceding virtual tab stop.
        chars = text.get("insert linestart", "insert")
        if chars == "":
            if text.compare("insert", ">", "1.0"):
                # easy: delete preceding newline
                text.delete("insert-1c")
            else:
                text.bell()  # at start of buffer
            return "break"
        if chars[-1] not in " \t":
            # easy: delete preceding real char
            text.delete("insert-1c")
            return "break"
        # Ick.  It may require *inserting* spaces if we back up over a
        # tab character!  This is written to be clear, not fast.
        have = len(chars.expandtabs(self.tabwidth))
        assert have > 0
        want = int((have - 1) / self.indentwidth) * self.indentwidth
        ncharsdeleted = 0
        while 1:
            chars = chars[:-1]
            ncharsdeleted = ncharsdeleted + 1
            have = len(chars.expandtabs(self.tabwidth))
            if have <= want or chars[-1] not in " \t":
                break
        text.undo_block_start()
        text.delete("insert-%dc" % ncharsdeleted, "insert")
        if have < want:
            text.insert("insert", " " * (want - have))
        text.undo_block_stop()
        return "break"

    def smart_indent_event(self, event):
        # if intraline selection:
        #     delete it
        # elif multiline selection:
        #     do indent-region & return
        # indent one level
        text = self.text
        first, last = self.editwin.get_selection_indices()
        text.undo_block_start()
        try:
            if first and last:
                if index2line(first) != index2line(last):
                    return self.indent_region_event(event)
                text.delete(first, last)
                text.mark_set("insert", first)
            prefix = text.get("insert linestart", "insert")
            raw, effective = classifyws(prefix, self.tabwidth)
            if raw == len(prefix):
                # only whitespace to the left
                self.reindent_to(effective + self.indentwidth)
            else:
                if self.usetabs:
                    pad = "\t"
                else:
                    effective = len(prefix.expandtabs(self.tabwidth))
                    n = self.indentwidth
                    pad = " " * (n - effective % n)
                text.insert("insert", pad)
            text.see("insert")
            return "break"
        finally:
            text.undo_block_stop()

    def newline_and_indent_event(self, event):
        text = self.text
        first, last = self.editwin.get_selection_indices()
        text.undo_block_start()
        try:
            if first and last:
                text.delete(first, last)
                text.mark_set("insert", first)
            line = text.get("insert linestart", "insert")
            i, n = 0, len(line)
            while i < n and line[i] in " \t":
                i = i + 1
            if i == n:
                # the cursor is in or at leading indentation; just inject
                # an empty line at the start and strip space from current line
                text.delete("insert - %d chars" % i, "insert")
                text.insert("insert linestart", "\n")
                return "break"
            indent = line[:i]
            # strip whitespace before insert point
            i = 0
            while line and line[-1] in " \t":
                line = line[:-1]
                i = i + 1
            if i:
                text.delete("insert - %d chars" % i, "insert")
            # strip whitespace after insert point
            while text.get("insert") in " \t":
                text.delete("insert")
            # start new line
            text.insert("insert", "\n")

            # adjust indentation for continuations and block
            # open/close first need to find the last stmt
            lno = index2line(text.index("insert"))
            y = PyParse.Parser(self.indentwidth, self.tabwidth)
            for context in self.num_context_lines:
                startat = max(lno - context, 1)
                startatindex = repr(startat) + ".0"
                rawtext = text.get(startatindex, "insert")
                y.set_str(rawtext)
                bod = y.find_good_parse_start(
                    self.context_use_ps1, self._build_char_in_string_func(startatindex)
                )
                if bod is not None or startat == 1:
                    break
            y.set_lo(bod or 0)
            c = y.get_continuation_type()
            if c != PyParse.C_NONE:
                # The current stmt hasn't ended yet.
                if c == PyParse.C_STRING:
                    # inside a string; just mimic the current indent
                    text.insert("insert", indent)
                elif c == PyParse.C_BRACKET:
                    # line up with the first (if any) element of the
                    # last open bracket structure; else indent one
                    # level beyond the indent of the line with the
                    # last open bracket
                    self.reindent_to(y.compute_bracket_indent())
                elif c == PyParse.C_BACKSLASH:
                    # if more than one line in this stmt already, just
                    # mimic the current indent; else if initial line
                    # has a start on an assignment stmt, indent to
                    # beyond leftmost =; else to beyond first chunk of
                    # non-whitespace on initial line
                    if y.get_num_lines_in_stmt() > 1:
                        text.insert("insert", indent)
                    else:
                        self.reindent_to(y.compute_backslash_indent())
                else:
                    assert 0, "bogus continuation type " + repr(c)
                return "break"

            # This line starts a brand new stmt; indent relative to
            # indentation of initial line of closest preceding
            # interesting stmt.
            indent = y.get_base_indent_string()
            text.insert("insert", indent)
            if y.is_block_opener():
                self.smart_indent_event(event)
            elif indent and y.is_block_closer():
                self.smart_backspace_event(event)
            return "break"
        finally:
            text.see("insert")
            text.undo_block_stop()

    auto_indent = newline_and_indent_event

    # Our editwin provides a is_char_in_string function that works
    # with a Tk text index, but PyParse only knows about offsets into
    # a string. This builds a function for PyParse that accepts an
    # offset.

    def _build_char_in_string_func(self, startindex):
        def inner(offset, _startindex=startindex, _icis=self.editwin.is_char_in_string):
            return _icis(_startindex + "+%dc" % offset)

        return inner

    def indent_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        for pos in range(len(lines)):
            line = lines[pos]
            if line:
                raw, effective = classifyws(line, self.tabwidth)
                effective = effective + self.indentwidth
                lines[pos] = self._make_blanks(effective) + line[raw:]
        self.set_region(head, tail, chars, lines)
        return "break"

    def dedent_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        for pos in range(len(lines)):
            line = lines[pos]
            if line:
                raw, effective = classifyws(line, self.tabwidth)
                effective = max(effective - self.indentwidth, 0)
                lines[pos] = self._make_blanks(effective) + line[raw:]
        self.set_region(head, tail, chars, lines)
        return "break"

    def comment_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        for pos in range(len(lines) - 1):
            line = lines[pos]
            lines[pos] = "##" + line
        self.set_region(head, tail, chars, lines)

    def uncomment_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        for pos in range(len(lines)):
            line = lines[pos]
            if not line:
                continue
            if line[:2] == "##":
                line = line[2:]
            elif line[:1] == "#":
                line = line[1:]
            lines[pos] = line
        self.set_region(head, tail, chars, lines)

    def tabify_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        tabwidth = self._asktabwidth()
        for pos in range(len(lines)):
            line = lines[pos]
            if line:
                raw, effective = classifyws(line, tabwidth)
                ntabs, nspaces = divmod(effective, tabwidth)
                lines[pos] = "\t" * ntabs + " " * nspaces + line[raw:]
        self.set_region(head, tail, chars, lines)

    def untabify_region_event(self, event):
        head, tail, chars, lines = self.get_region()
        tabwidth = self._asktabwidth()
        for pos in range(len(lines)):
            lines[pos] = lines[pos].expandtabs(tabwidth)
        self.set_region(head, tail, chars, lines)

    def toggle_tabs_event(self, event):
        if self.editwin.askyesno(
            "Toggle tabs",
            "Turn tabs " + ("on", "off")[self.usetabs] + "?",
            parent=self.text,
        ):
            self.usetabs = not self.usetabs
        return "break"

    # XXX this isn't bound to anything -- see class tabwidth comments
    def change_tabwidth_event(self, event):
        new = self._asktabwidth()
        if new != self.tabwidth:
            self.tabwidth = new
            self.set_indentation_params(0, guess=0)
        return "break"

    def change_indentwidth_event(self, event):
        new = self.editwin.askinteger(
            "Indent width",
            "New indent width (1-16)",
            parent=self.text,
            initialvalue=self.indentwidth,
            minvalue=1,
            maxvalue=16,
        )
        if new and new != self.indentwidth:
            self.indentwidth = new
        return "break"

    def get_region(self):
        text = self.text
        first, last = self.editwin.get_selection_indices()
        if first and last:
            head = text.index(first + " linestart")
            tail = text.index(last + "-1c lineend +1c")
        else:
            head = text.index("insert linestart")
            tail = text.index("insert lineend +1c")
        chars = text.get(head, tail)
        lines = chars.split("\n")
        return head, tail, chars, lines

    def set_region(self, head, tail, chars, lines):
        text = self.text
        newchars = "\n".join(lines)
        if newchars == chars:
            text.bell()
            return
        text.tag_remove("sel", "1.0", "end")
        text.mark_set("insert", head)
        text.undo_block_start()
        text.delete(head, tail)
        text.insert(head, newchars)
        text.undo_block_stop()
        text.tag_add("sel", head, "insert")

    # Make string that displays as n leading blanks.

    def _make_blanks(self, n):
        if self.usetabs:
            ntabs, nspaces = divmod(n, self.tabwidth)
            return "\t" * ntabs + " " * nspaces
        else:
            return " " * n

    # Delete from beginning of line to insert point, then reinsert
    # column logical (meaning use tabs if appropriate) spaces.

    def reindent_to(self, column):
        text = self.text
        text.undo_block_start()
        if text.compare("insert linestart", "!=", "insert"):
            text.delete("insert linestart", "insert")
        if column:
            text.insert("insert", self._make_blanks(column))
        text.undo_block_stop()

    def _asktabwidth(self):
        return (
            self.editwin.askinteger(
                "Tab width",
                "Spaces per tab?",
                parent=self.text,
                initialvalue=self.tabwidth,
                minvalue=1,
                maxvalue=16,
            )
            or self.tabwidth
        )

    # Guess indentwidth from text content.
    # Return guessed indentwidth.  This should not be believed unless
    # it's in a reasonable range (e.g., it will be 0 if no indented
    # blocks are found).

    def guess_indent(self):
        opener, indented = IndentSearcher(self.text, self.tabwidth).run()
        if opener and indented:
            raw, indentsmall = classifyws(opener, self.tabwidth)
            raw, indentlarge = classifyws(indented, self.tabwidth)
        else:
            indentsmall = indentlarge = 0
        return indentlarge - indentsmall


# "line.col" -> line, as an int
def index2line(index):
    return int(float(index))


# Look at the leading whitespace in s.
# Return pair (# of leading ws characters,
#              effective # of leading blanks after expanding
#              tabs to width tabwidth)


def classifyws(s, tabwidth):
    raw = effective = 0
    for ch in s:
        if ch == " ":
            raw = raw + 1
            effective = effective + 1
        elif ch == "\t":
            raw = raw + 1
            effective = (effective // tabwidth + 1) * tabwidth
        else:
            break
    return raw, effective


class IndentSearcher:

    # .run() chews over the Text widget, looking for a block opener
    # and the stmt following it.  Returns a pair,
    #     (line containing block opener, line containing stmt)
    # Either or both may be None.

    def __init__(self, text, tabwidth):
        self.text = text
        self.tabwidth = tabwidth
        self.i = self.finished = 0
        self.blkopenline = self.indentedline = None

    def readline(self):
        if self.finished:
            val = ""
        else:
            i = self.i = self.i + 1
            mark = repr(i) + ".0"
            if self.text.compare(mark, ">=", "end"):
                val = ""
            else:
                val = self.text.get(mark, mark + " lineend+1c")
        # hrm - not sure this is correct in py3k - the source code may have
        # an encoding declared, but the data will *always* be in
        # default_scintilla_encoding - so if anyone looks at the encoding decl
        # in the source they will be wrong.  I think.  Maybe.  Or something...
        return val.encode(default_scintilla_encoding)

    def run(self):
        OPENERS = ("class", "def", "for", "if", "try", "while")
        INDENT = tokenize.INDENT
        NAME = tokenize.NAME

        save_tabsize = tokenize.tabsize
        tokenize.tabsize = self.tabwidth
        try:
            try:
                for (typ, token, start, end, line) in token_generator(self.readline):
                    if typ == NAME and token in OPENERS:
                        self.blkopenline = line
                    elif typ == INDENT and self.blkopenline:
                        self.indentedline = line
                        break

            except (tokenize.TokenError, IndentationError):
                # since we cut off the tokenizer early, we can trigger
                # spurious errors
                pass
        finally:
            tokenize.tabsize = save_tabsize
        return self.blkopenline, self.indentedline



import win32api
import win32wnet
import sys
from winnetwk import *
import os

possible_shares = []


def _doDumpHandle(handle, level=0):
    indent = " " * level
    while 1:
        items = win32wnet.WNetEnumResource(handle, 0)
        if len(items) == 0:
            break
        for item in items:
            try:
                if item.dwDisplayType == RESOURCEDISPLAYTYPE_SHARE:
                    print(indent + "Have share with name:", item.lpRemoteName)
                    possible_shares.append(item)
                elif item.dwDisplayType == RESOURCEDISPLAYTYPE_GENERIC:
                    print(
                        indent + "Have generic resource with name:", item.lpRemoteName
                    )
                else:
                    # Try generic!
                    print(indent + "Enumerating " + item.lpRemoteName, end=" ")
                    k = win32wnet.WNetOpenEnum(
                        RESOURCE_GLOBALNET, RESOURCETYPE_ANY, 0, item
                    )
                    print()
                    _doDumpHandle(k, level + 1)
                    win32wnet.WNetCloseEnum(
                        k
                    )  # could do k.Close(), but this is a good test!
            except win32wnet.error as details:
                print(indent + "Couldn't enumerate this resource: " + details.strerror)


def TestOpenEnum():
    print("Enumerating all resources on the network - this may take some time...")
    handle = win32wnet.WNetOpenEnum(RESOURCE_GLOBALNET, RESOURCETYPE_ANY, 0, None)

    try:
        _doDumpHandle(handle)
    finally:
        handle.Close()
    print("Finished dumping all resources.")


def findUnusedDriveLetter():
    existing = [
        x[0].lower() for x in win32api.GetLogicalDriveStrings().split("\0") if x
    ]
    handle = win32wnet.WNetOpenEnum(RESOURCE_REMEMBERED, RESOURCETYPE_DISK, 0, None)
    try:
        while 1:
            items = win32wnet.WNetEnumResource(handle, 0)
            if len(items) == 0:
                break
            xtra = [i.lpLocalName[0].lower() for i in items if i.lpLocalName]
            existing.extend(xtra)
    finally:
        handle.Close()
    for maybe in "defghijklmnopqrstuvwxyz":
        if maybe not in existing:
            return maybe
    raise RuntimeError("All drive mappings are taken?")


def TestConnection():
    if len(possible_shares) == 0:
        print("Couldn't find any potential shares to connect to")
        return
    localName = findUnusedDriveLetter() + ":"
    for share in possible_shares:
        print("Attempting connection of", localName, "to", share.lpRemoteName)
        try:
            win32wnet.WNetAddConnection2(share.dwType, localName, share.lpRemoteName)
        except win32wnet.error as details:
            print("Couldn't connect: " + details.strerror)
            continue
        # Have a connection.
        try:
            fname = os.path.join(localName + "\\", os.listdir(localName + "\\")[0])
            try:
                print(
                    "Universal name of '%s' is '%s'"
                    % (fname, win32wnet.WNetGetUniversalName(fname))
                )
            except win32wnet.error as details:
                print(
                    "Couldn't get universal name of '%s': %s"
                    % (fname, details.strerror)
                )
            print("User name for this connection is", win32wnet.WNetGetUser(localName))
        finally:
            win32wnet.WNetCancelConnection2(localName, 0, 0)
        # and do it again, but this time by using the more modern
        # NETRESOURCE way.
        nr = win32wnet.NETRESOURCE()
        nr.dwType = share.dwType
        nr.lpLocalName = localName
        nr.lpRemoteName = share.lpRemoteName
        win32wnet.WNetAddConnection2(nr)
        win32wnet.WNetCancelConnection2(localName, 0, 0)

        # and one more time using WNetAddConnection3
        win32wnet.WNetAddConnection3(0, nr)
        win32wnet.WNetCancelConnection2(localName, 0, 0)

        # Only do the first share that succeeds.
        break


def TestGetUser():
    u = win32wnet.WNetGetUser()
    print("Current global user is", repr(u))
    if u != win32wnet.WNetGetUser(None):
        raise RuntimeError("Default value didnt seem to work!")


TestGetUser()
TestOpenEnum()
TestConnection()



##
## helloapp.py
##
##
## A nice, small 'hello world' Pythonwin application.
## NOT an MDI application - just a single, normal, top-level window.
##
## MUST be run with the command line "pythonwin.exe /app helloapp.py"
## (or if you are really keen, rename "pythonwin.exe" to something else, then
## using MSVC or similar, edit the string section in the .EXE to name this file)
##
## Originally by Willy Heineman <wheineman@uconect.net>


import win32con
import win32ui
from pywin.mfc import window, dialog, afxres
from pywin.mfc.thread import WinApp

# The main frame.
# Does almost nothing at all - doesnt even create a child window!
class HelloWindow(window.Wnd):
    def __init__(self):
        # The window.Wnd ctor creates a Window object, and places it in
        # self._obj_.  Note the window object exists, but the window itself
        # does not!
        window.Wnd.__init__(self, win32ui.CreateWnd())

        # Now we ask the window object to create the window itself.
        self._obj_.CreateWindowEx(
            win32con.WS_EX_CLIENTEDGE,
            win32ui.RegisterWndClass(0, 0, win32con.COLOR_WINDOW + 1),
            "Hello World!",
            win32con.WS_OVERLAPPEDWINDOW,
            (100, 100, 400, 300),
            None,
            0,
            None,
        )


# The application object itself.
class HelloApp(WinApp):
    def InitInstance(self):
        self.frame = HelloWindow()
        self.frame.ShowWindow(win32con.SW_SHOWNORMAL)
        # We need to tell MFC what our main frame is.
        self.SetMainFrame(self.frame)


# Now create the application object itself!
app = HelloApp()



#
# bulkstamp.py:
#    Stamp versions on all files that can be found in a given tree.
#
# USAGE: python bulkstamp.py <version> <root directory> <descriptions>
#
# Example: python bulkstamp.py 103 ..\win32\Build\ desc.txt
#
# <version> corresponds to the build number. It will be concatenated with
# the major and minor version numbers found in the description file.
#
# Description information is pulled from an input text file with lines of
# the form:
#
#    <basename> <white space> <description>
#
# For example:
#
#    PyWinTypes.dll Common types for Python on Win32
#    etc
#
# The product's name, major, and minor versions are specified as:
#
#    name <white space> <value>
#    major <white space> <value>
#    minor <white space> <value>
#
# The tags are case-sensitive.
#
# Any line beginning with "#" will be ignored. Empty lines are okay.
#

import sys
import os
import verstamp
import fnmatch
import win32api

numStamped = 0

g_patterns = [
    "*.dll",
    "*.pyd",
    "*.exe",
    "*.ocx",
]


def walk(arg, dirname, names):
    global numStamped
    vars, debug, descriptions = arg
    for name in names:
        for pat in g_patterns:
            if fnmatch.fnmatch(name, pat):
                # Handle the "_d" thing.
                pathname = os.path.join(dirname, name)
                base, ext = os.path.splitext(name)
                if base[-2:] == "_d":
                    name = base[:-2] + ext
                is_dll = ext.lower() != ".exe"
                if os.path.normcase(name) in descriptions:
                    desc = descriptions[os.path.normcase(name)]
                    try:
                        verstamp.stamp(vars, pathname, desc, is_dll=is_dll)
                        numStamped = numStamped + 1
                    except win32api.error as exc:
                        print(
                            "Could not stamp",
                            pathname,
                            "Error",
                            exc.winerror,
                            "-",
                            exc.strerror,
                        )
                else:
                    print("WARNING: description not provided for:", name)
                    # skip branding this - assume already branded or handled elsewhere


#        print "Stamped", pathname


def load_descriptions(fname, vars):
    retvars = {}
    descriptions = {}

    lines = open(fname, "r").readlines()

    for i in range(len(lines)):
        line = lines[i].strip()
        if line != "" and line[0] != "#":
            idx1 = line.find(" ")
            idx2 = line.find("\t")
            if idx1 == -1 or idx2 < idx1:
                idx1 = idx2
            if idx1 == -1:
                print("ERROR: bad syntax in description file at line %d." % (i + 1))
                sys.exit(1)

            key = line[:idx1]
            val = line[idx1:].strip()
            if key in vars:
                retvars[key] = val
            else:
                descriptions[key] = val

    if "product" not in retvars:
        print("ERROR: description file is missing the product name.")
        sys.exit(1)
    if "major" not in retvars:
        print("ERROR: description file is missing the major version number.")
        sys.exit(1)
    if "minor" not in retvars:
        print("ERROR: description file is missing the minor version number.")
        sys.exit(1)

    return retvars, descriptions


def scan(build, root, desc, **custom_vars):
    global numStamped
    numStamped = 0
    try:
        build = int(build)
    except ValueError:
        print("ERROR: build number is not a number: %s" % build)
        sys.exit(1)

    debug = 0  ### maybe fix this one day

    varList = ["major", "minor", "sub", "company", "copyright", "trademarks", "product"]

    vars, descriptions = load_descriptions(desc, varList)
    vars["build"] = build
    vars.update(custom_vars)

    arg = vars, debug, descriptions
    os.path.walk(root, walk, arg)

    print("Stamped %d files." % (numStamped))


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("ERROR: incorrect invocation. See script's header comments.")
        sys.exit(1)

    scan(*tuple(sys.argv[1:]))



from win32com.client import Dispatch, DispatchWithEvents, constants
import pythoncom
import os
import time

finished = 0  # Flag for the wait loop from (3) to test


class ADOEvents:  # event handler class
    def OnWillConnect(self, str, user, pw, opt, sts, cn):
        # Must have this event, as if it is not handled, ADO assumes the
        # operation is cancelled, and raises an error (Operation cancelled
        # by the user)
        pass

    def OnConnectComplete(self, error, status, connection):
        # Assume no errors, until we have the basic stuff
        # working. Now, "connection" should be an open
        # connection to my data source
        # Do the "something" from (2). For now, just
        # print the connection data source
        print("connection is", connection)
        print("Connected to", connection.Properties("Data Source"))
        # OK, our work is done. Let the main loop know
        global finished
        finished = 1

    def OnCommitTransComplete(self, pError, adStatus, pConnection):
        pass

    def OnInfoMessage(self, pError, adStatus, pConnection):
        pass

    def OnDisconnect(self, adStatus, pConnection):
        pass

    def OnBeginTransComplete(self, TransactionLevel, pError, adStatus, pConnection):
        pass

    def OnRollbackTransComplete(self, pError, adStatus, pConnection):
        pass

    def OnExecuteComplete(
        self, RecordsAffected, pError, adStatus, pCommand, pRecordset, pConnection
    ):
        pass

    def OnWillExecute(
        self,
        Source,
        CursorType,
        LockType,
        Options,
        adStatus,
        pCommand,
        pRecordset,
        pConnection,
    ):
        pass


def TestConnection(dbname):
    # Create the ADO connection object, and link the event
    # handlers into it
    c = DispatchWithEvents("ADODB.Connection", ADOEvents)

    # Initiate the asynchronous open
    dsn = "Driver={Microsoft Access Driver (*.mdb)};Dbq=%s" % dbname
    user = "system"
    pw = "manager"
    c.Open(dsn, user, pw, constants.adAsyncConnect)

    # Sit in a loop, until our event handler (above) sets the
    # "finished" flag or we time out.
    end_time = time.clock() + 10
    while time.clock() < end_time:
        # Pump messages so that COM gets a look in
        pythoncom.PumpWaitingMessages()
    if not finished:
        print("XXX - Failed to connect!")


def Test():
    from . import testAccess

    try:
        testAccess.GenerateSupport()
    except pythoncom.com_error:
        print("*** Can not import the MSAccess type libraries - tests skipped")
        return
    dbname = testAccess.CreateTestAccessDatabase()
    try:
        TestConnection(dbname)
    finally:
        os.unlink(dbname)


if __name__ == "__main__":
    Test()



if type(__path__) == type(""):
    # For freeze to work!
    import sys

    try:
        import mapi

        sys.modules["win32com.mapi.mapi"] = mapi
    except ImportError:
        pass
    try:
        import exchange

        sys.modules["win32com.mapi.exchange"] = exchange
    except ImportError:
        pass
    try:
        import exchdapi

        sys.modules["win32com.mapi.exchdapi"] = exchdapi
    except ImportError:
        pass
else:
    import win32com

    # See if we have a special directory for the binaries (for developers)
    win32com.__PackageSupportBuildPath__(__path__)



"""genpy.py - The worker for makepy.  See makepy.py for more details

This code was moved simply to speed Python in normal circumstances.  As the makepy.py
is normally run from the command line, it reparses the code each time.  Now makepy
is nothing more than the command line handler and public interface.

The makepy command line etc handling is also getting large enough in its own right!
"""

# NOTE - now supports a "demand" mechanism - the top-level is a package, and
# each class etc can be made individually.
# This should eventually become the default.
# Then the old non-package technique should be removed.
# There should be no b/w compat issues, and will just help clean the code.
# This will be done once the new "demand" mechanism gets a good workout.
import os
import sys
import time
import win32com

import pythoncom
from . import build

error = "makepy.error"
makepy_version = "0.5.01"  # Written to generated file.

GEN_FULL = "full"
GEN_DEMAND_BASE = "demand(base)"
GEN_DEMAND_CHILD = "demand(child)"

# This map is used purely for the users benefit -it shows the
# raw, underlying type of Alias/Enums, etc.  The COM implementation
# does not use this map at runtime - all Alias/Enum have already
# been translated.
mapVTToTypeString = {
    pythoncom.VT_I2: "types.IntType",
    pythoncom.VT_I4: "types.IntType",
    pythoncom.VT_R4: "types.FloatType",
    pythoncom.VT_R8: "types.FloatType",
    pythoncom.VT_BSTR: "types.StringType",
    pythoncom.VT_BOOL: "types.IntType",
    pythoncom.VT_VARIANT: "types.TypeType",
    pythoncom.VT_I1: "types.IntType",
    pythoncom.VT_UI1: "types.IntType",
    pythoncom.VT_UI2: "types.IntType",
    pythoncom.VT_UI4: "types.IntType",
    pythoncom.VT_I8: "types.LongType",
    pythoncom.VT_UI8: "types.LongType",
    pythoncom.VT_INT: "types.IntType",
    pythoncom.VT_DATE: "pythoncom.PyTimeType",
    pythoncom.VT_UINT: "types.IntType",
}

# Given a propget function's arg desc, return the default parameters for all
# params bar the first.  Eg, then Python does a:
# object.Property = "foo"
# Python can only pass the "foo" value.  If the property has
# multiple args, and the rest have default values, this allows
# Python to correctly pass those defaults.
def MakeDefaultArgsForPropertyPut(argsDesc):
    ret = []
    for desc in argsDesc[1:]:
        default = build.MakeDefaultArgRepr(desc)
        if default is None:
            break
        ret.append(default)
    return tuple(ret)


def MakeMapLineEntry(dispid, wFlags, retType, argTypes, user, resultCLSID):
    # Strip the default value
    argTypes = tuple([what[:2] for what in argTypes])
    return '(%s, %d, %s, %s, "%s", %s)' % (
        dispid,
        wFlags,
        retType[:2],
        argTypes,
        user,
        resultCLSID,
    )


def MakeEventMethodName(eventName):
    if eventName[:2] == "On":
        return eventName
    else:
        return "On" + eventName


def WriteSinkEventMap(obj, stream):
    print("\t_dispid_to_func_ = {", file=stream)
    for name, entry in (
        list(obj.propMapGet.items())
        + list(obj.propMapPut.items())
        + list(obj.mapFuncs.items())
    ):
        fdesc = entry.desc
        print(
            '\t\t%9d : "%s",' % (fdesc.memid, MakeEventMethodName(entry.names[0])),
            file=stream,
        )
    print("\t\t}", file=stream)


# MI is used to join my writable helpers, and the OLE
# classes.
class WritableItem:
    # __cmp__ used for sorting in py2x...
    def __cmp__(self, other):
        "Compare for sorting"
        ret = cmp(self.order, other.order)
        if ret == 0 and self.doc:
            ret = cmp(self.doc[0], other.doc[0])
        return ret

    # ... but not used in py3k - __lt__ minimum needed there
    def __lt__(self, other):  # py3k variant
        if self.order == other.order:
            return self.doc < other.doc
        return self.order < other.order

    def __repr__(self):
        return "OleItem: doc=%s, order=%d" % (repr(self.doc), self.order)


class RecordItem(build.OleItem, WritableItem):
    order = 9
    typename = "RECORD"

    def __init__(self, typeInfo, typeAttr, doc=None, bForUser=1):
        ##    sys.stderr.write("Record %s: size %s\n" % (doc,typeAttr.cbSizeInstance))
        ##    sys.stderr.write(" cVars = %s\n" % (typeAttr.cVars,))
        ##    for i in range(typeAttr.cVars):
        ##        vdesc = typeInfo.GetVarDesc(i)
        ##        sys.stderr.write(" Var %d has value %s, type %d, desc=%s\n" % (i, vdesc.value, vdesc.varkind, vdesc.elemdescVar))
        ##        sys.stderr.write(" Doc is %s\n" % (typeInfo.GetDocumentation(vdesc.memid),))

        build.OleItem.__init__(self, doc)
        self.clsid = typeAttr[0]

    def WriteClass(self, generator):
        pass


# Given an enum, write all aliases for it.
# (no longer necessary for new style code, but still used for old code.
def WriteAliasesForItem(item, aliasItems, stream):
    for alias in aliasItems.values():
        if item.doc and alias.aliasDoc and (alias.aliasDoc[0] == item.doc[0]):
            alias.WriteAliasItem(aliasItems, stream)


class AliasItem(build.OleItem, WritableItem):
    order = 2
    typename = "ALIAS"

    def __init__(self, typeinfo, attr, doc=None, bForUser=1):
        build.OleItem.__init__(self, doc)

        ai = attr[14]
        self.attr = attr
        if type(ai) == type(()) and type(ai[1]) == type(
            0
        ):  # XXX - This is a hack - why tuples?  Need to resolve?
            href = ai[1]
            alinfo = typeinfo.GetRefTypeInfo(href)
            self.aliasDoc = alinfo.GetDocumentation(-1)
            self.aliasAttr = alinfo.GetTypeAttr()
        else:
            self.aliasDoc = None
            self.aliasAttr = None

    def WriteAliasItem(self, aliasDict, stream):
        # we could have been written as part of an alias dependency
        if self.bWritten:
            return

        if self.aliasDoc:
            depName = self.aliasDoc[0]
            if depName in aliasDict:
                aliasDict[depName].WriteAliasItem(aliasDict, stream)
            print(self.doc[0] + " = " + depName, file=stream)
        else:
            ai = self.attr[14]
            if type(ai) == type(0):
                try:
                    typeStr = mapVTToTypeString[ai]
                    print("# %s=%s" % (self.doc[0], typeStr), file=stream)
                except KeyError:
                    print(
                        self.doc[0] + " = None # Can't convert alias info " + str(ai),
                        file=stream,
                    )
        print(file=stream)
        self.bWritten = 1


class EnumerationItem(build.OleItem, WritableItem):
    order = 1
    typename = "ENUMERATION"

    def __init__(self, typeinfo, attr, doc=None, bForUser=1):
        build.OleItem.__init__(self, doc)

        self.clsid = attr[0]
        self.mapVars = {}
        typeFlags = attr[11]
        self.hidden = (
            typeFlags & pythoncom.TYPEFLAG_FHIDDEN
            or typeFlags & pythoncom.TYPEFLAG_FRESTRICTED
        )

        for j in range(attr[7]):
            vdesc = typeinfo.GetVarDesc(j)
            name = typeinfo.GetNames(vdesc[0])[0]
            self.mapVars[name] = build.MapEntry(vdesc)

    ##  def WriteEnumerationHeaders(self, aliasItems, stream):
    ##    enumName = self.doc[0]
    ##    print >> stream "%s=constants # Compatibility with previous versions." % (enumName)
    ##    WriteAliasesForItem(self, aliasItems)

    def WriteEnumerationItems(self, stream):
        num = 0
        enumName = self.doc[0]
        # Write in name alpha order
        names = list(self.mapVars.keys())
        names.sort()
        for name in names:
            entry = self.mapVars[name]
            vdesc = entry.desc
            if vdesc[4] == pythoncom.VAR_CONST:
                val = vdesc[1]

                use = repr(val)
                # Make sure the repr of the value is valid python syntax
                # still could cause an error on import if it contains a module or type name
                # not available in the global namespace
                try:
                    compile(use, "<makepy>", "eval")
                except SyntaxError:
                    # At least add the repr as a string, so it can be investigated further
                    # Sanitize it, in case the repr contains its own quotes.  (??? line breaks too ???)
                    use = use.replace('"', "'")
                    use = (
                        '"'
                        + use
                        + '"'
                        + " # This VARIANT type cannot be converted automatically"
                    )
                print(
                    "\t%-30s=%-10s # from enum %s"
                    % (build.MakePublicAttributeName(name, True), use, enumName),
                    file=stream,
                )
                num += 1
        return num


class VTableItem(build.VTableItem, WritableItem):
    order = 4

    def WriteClass(self, generator):
        self.WriteVTableMap(generator)
        self.bWritten = 1

    def WriteVTableMap(self, generator):
        stream = generator.file
        print(
            "%s_vtables_dispatch_ = %d" % (self.python_name, self.bIsDispatch),
            file=stream,
        )
        print("%s_vtables_ = [" % (self.python_name,), file=stream)
        for v in self.vtableFuncs:
            names, dispid, desc = v
            assert desc.desckind == pythoncom.DESCKIND_FUNCDESC
            arg_reprs = []
            # more hoops so we don't generate huge lines.
            item_num = 0
            print("\t((", end=" ", file=stream)
            for name in names:
                print(repr(name), ",", end=" ", file=stream)
                item_num = item_num + 1
                if item_num % 5 == 0:
                    print("\n\t\t\t", end=" ", file=stream)
            print(
                "), %d, (%r, %r, [" % (dispid, desc.memid, desc.scodeArray),
                end=" ",
                file=stream,
            )
            for arg in desc.args:
                item_num = item_num + 1
                if item_num % 5 == 0:
                    print("\n\t\t\t", end=" ", file=stream)
                defval = build.MakeDefaultArgRepr(arg)
                if arg[3] is None:
                    arg3_repr = None
                else:
                    arg3_repr = repr(arg[3])
                print(
                    repr((arg[0], arg[1], defval, arg3_repr)), ",", end=" ", file=stream
                )
            print("],", end=" ", file=stream)
            print(repr(desc.funckind), ",", end=" ", file=stream)
            print(repr(desc.invkind), ",", end=" ", file=stream)
            print(repr(desc.callconv), ",", end=" ", file=stream)
            print(repr(desc.cParamsOpt), ",", end=" ", file=stream)
            print(repr(desc.oVft), ",", end=" ", file=stream)
            print(repr(desc.rettype), ",", end=" ", file=stream)
            print(repr(desc.wFuncFlags), ",", end=" ", file=stream)
            print(")),", file=stream)
        print("]", file=stream)
        print(file=stream)


class DispatchItem(build.DispatchItem, WritableItem):
    order = 3

    def __init__(self, typeinfo, attr, doc=None):
        build.DispatchItem.__init__(self, typeinfo, attr, doc)
        self.type_attr = attr
        self.coclass_clsid = None

    def WriteClass(self, generator):
        if (
            not self.bIsDispatch
            and not self.type_attr.typekind == pythoncom.TKIND_DISPATCH
        ):
            return
        # This is pretty screwey - now we have vtable support we
        # should probably rethink this (ie, maybe write both sides for sinks, etc)
        if self.bIsSink:
            self.WriteEventSinkClassHeader(generator)
            self.WriteCallbackClassBody(generator)
        else:
            self.WriteClassHeader(generator)
            self.WriteClassBody(generator)
        print(file=generator.file)
        self.bWritten = 1

    def WriteClassHeader(self, generator):
        generator.checkWriteDispatchBaseClass()
        doc = self.doc
        stream = generator.file
        print("class " + self.python_name + "(DispatchBaseClass):", file=stream)
        if doc[1]:
            print("\t" + build._makeDocString(doc[1]), file=stream)
        try:
            progId = pythoncom.ProgIDFromCLSID(self.clsid)
            print(
                "\t# This class is creatable by the name '%s'" % (progId), file=stream
            )
        except pythoncom.com_error:
            pass
        print("\tCLSID = " + repr(self.clsid), file=stream)
        if self.coclass_clsid is None:
            print("\tcoclass_clsid = None", file=stream)
        else:
            print("\tcoclass_clsid = " + repr(self.coclass_clsid), file=stream)
        print(file=stream)
        self.bWritten = 1

    def WriteEventSinkClassHeader(self, generator):
        generator.checkWriteEventBaseClass()
        doc = self.doc
        stream = generator.file
        print("class " + self.python_name + ":", file=stream)
        if doc[1]:
            print("\t" + build._makeDocString(doc[1]), file=stream)
        try:
            progId = pythoncom.ProgIDFromCLSID(self.clsid)
            print(
                "\t# This class is creatable by the name '%s'" % (progId), file=stream
            )
        except pythoncom.com_error:
            pass
        print("\tCLSID = CLSID_Sink = " + repr(self.clsid), file=stream)
        if self.coclass_clsid is None:
            print("\tcoclass_clsid = None", file=stream)
        else:
            print("\tcoclass_clsid = " + repr(self.coclass_clsid), file=stream)
        print("\t_public_methods_ = [] # For COM Server support", file=stream)
        WriteSinkEventMap(self, stream)
        print(file=stream)
        print("\tdef __init__(self, oobj = None):", file=stream)
        print("\t\tif oobj is None:", file=stream)
        print("\t\t\tself._olecp = None", file=stream)
        print("\t\telse:", file=stream)
        print("\t\t\timport win32com.server.util", file=stream)
        print(
            "\t\t\tfrom win32com.server.policy import EventHandlerPolicy", file=stream
        )
        print(
            "\t\t\tcpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)",
            file=stream,
        )
        print("\t\t\tcp=cpc.FindConnectionPoint(self.CLSID_Sink)", file=stream)
        print(
            "\t\t\tcookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))",
            file=stream,
        )
        print("\t\t\tself._olecp,self._olecp_cookie = cp,cookie", file=stream)
        print("\tdef __del__(self):", file=stream)
        print("\t\ttry:", file=stream)
        print("\t\t\tself.close()", file=stream)
        print("\t\texcept pythoncom.com_error:", file=stream)
        print("\t\t\tpass", file=stream)
        print("\tdef close(self):", file=stream)
        print("\t\tif self._olecp is not None:", file=stream)
        print(
            "\t\t\tcp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None",
            file=stream,
        )
        print("\t\t\tcp.Unadvise(cookie)", file=stream)
        print("\tdef _query_interface_(self, iid):", file=stream)
        print("\t\timport win32com.server.util", file=stream)
        print(
            "\t\tif iid==self.CLSID_Sink: return win32com.server.util.wrap(self)",
            file=stream,
        )
        print(file=stream)
        self.bWritten = 1

    def WriteCallbackClassBody(self, generator):
        stream = generator.file
        print("\t# Event Handlers", file=stream)
        print(
            "\t# If you create handlers, they should have the following prototypes:",
            file=stream,
        )
        for name, entry in (
            list(self.propMapGet.items())
            + list(self.propMapPut.items())
            + list(self.mapFuncs.items())
        ):
            fdesc = entry.desc
            methName = MakeEventMethodName(entry.names[0])
            print(
                "#\tdef "
                + methName
                + "(self"
                + build.BuildCallList(
                    fdesc,
                    entry.names,
                    "defaultNamedOptArg",
                    "defaultNamedNotOptArg",
                    "defaultUnnamedArg",
                    "pythoncom.Missing",
                    is_comment=True,
                )
                + "):",
                file=stream,
            )
            if entry.doc and entry.doc[1]:
                print("#\t\t" + build._makeDocString(entry.doc[1]), file=stream)
        print(file=stream)
        self.bWritten = 1

    def WriteClassBody(self, generator):
        stream = generator.file
        # Write in alpha order.
        names = list(self.mapFuncs.keys())
        names.sort()
        specialItems = {
            "count": None,
            "item": None,
            "value": None,
            "_newenum": None,
        }  # If found, will end up with (entry, invoke_tupe)
        itemCount = None
        for name in names:
            entry = self.mapFuncs[name]
            assert entry.desc.desckind == pythoncom.DESCKIND_FUNCDESC
            # skip [restricted] methods, unless it is the
            # enumerator (which, being part of the "system",
            # we know about and can use)
            dispid = entry.desc.memid
            if (
                entry.desc.wFuncFlags & pythoncom.FUNCFLAG_FRESTRICTED
                and dispid != pythoncom.DISPID_NEWENUM
            ):
                continue
            # If not accessible via IDispatch, then we can't use it here.
            if entry.desc.funckind != pythoncom.FUNC_DISPATCH:
                continue
            if dispid == pythoncom.DISPID_VALUE:
                lkey = "value"
            elif dispid == pythoncom.DISPID_NEWENUM:
                specialItems["_newenum"] = (entry, entry.desc.invkind, None)
                continue  # Dont build this one now!
            else:
                lkey = name.lower()
            if (
                lkey in specialItems and specialItems[lkey] is None
            ):  # remember if a special one.
                specialItems[lkey] = (entry, entry.desc.invkind, None)
            if generator.bBuildHidden or not entry.hidden:
                if entry.GetResultName():
                    print("\t# Result is of type " + entry.GetResultName(), file=stream)
                if entry.wasProperty:
                    print(
                        "\t# The method %s is actually a property, but must be used as a method to correctly pass the arguments"
                        % name,
                        file=stream,
                    )
                ret = self.MakeFuncMethod(entry, build.MakePublicAttributeName(name))
                for line in ret:
                    print(line, file=stream)
        print("\t_prop_map_get_ = {", file=stream)
        names = list(self.propMap.keys())
        names.sort()
        for key in names:
            entry = self.propMap[key]
            if generator.bBuildHidden or not entry.hidden:
                resultName = entry.GetResultName()
                if resultName:
                    print(
                        "\t\t# Property '%s' is an object of type '%s'"
                        % (key, resultName),
                        file=stream,
                    )
                lkey = key.lower()
                details = entry.desc
                resultDesc = details[2]
                argDesc = ()
                mapEntry = MakeMapLineEntry(
                    details.memid,
                    pythoncom.DISPATCH_PROPERTYGET,
                    resultDesc,
                    argDesc,
                    key,
                    entry.GetResultCLSIDStr(),
                )

                if details.memid == pythoncom.DISPID_VALUE:
                    lkey = "value"
                elif details.memid == pythoncom.DISPID_NEWENUM:
                    lkey = "_newenum"
                else:
                    lkey = key.lower()
                if (
                    lkey in specialItems and specialItems[lkey] is None
                ):  # remember if a special one.
                    specialItems[lkey] = (
                        entry,
                        pythoncom.DISPATCH_PROPERTYGET,
                        mapEntry,
                    )
                    # All special methods, except _newenum, are written
                    # "normally".  This is a mess!
                    if details.memid == pythoncom.DISPID_NEWENUM:
                        continue

                print(
                    '\t\t"%s": %s,' % (build.MakePublicAttributeName(key), mapEntry),
                    file=stream,
                )
        names = list(self.propMapGet.keys())
        names.sort()
        for key in names:
            entry = self.propMapGet[key]
            if generator.bBuildHidden or not entry.hidden:
                if entry.GetResultName():
                    print(
                        "\t\t# Method '%s' returns object of type '%s'"
                        % (key, entry.GetResultName()),
                        file=stream,
                    )
                details = entry.desc
                assert details.desckind == pythoncom.DESCKIND_FUNCDESC
                lkey = key.lower()
                argDesc = details[2]
                resultDesc = details[8]
                mapEntry = MakeMapLineEntry(
                    details[0],
                    pythoncom.DISPATCH_PROPERTYGET,
                    resultDesc,
                    argDesc,
                    key,
                    entry.GetResultCLSIDStr(),
                )
                if details.memid == pythoncom.DISPID_VALUE:
                    lkey = "value"
                elif details.memid == pythoncom.DISPID_NEWENUM:
                    lkey = "_newenum"
                else:
                    lkey = key.lower()
                if (
                    lkey in specialItems and specialItems[lkey] is None
                ):  # remember if a special one.
                    specialItems[lkey] = (
                        entry,
                        pythoncom.DISPATCH_PROPERTYGET,
                        mapEntry,
                    )
                    # All special methods, except _newenum, are written
                    # "normally".  This is a mess!
                    if details.memid == pythoncom.DISPID_NEWENUM:
                        continue
                print(
                    '\t\t"%s": %s,' % (build.MakePublicAttributeName(key), mapEntry),
                    file=stream,
                )

        print("\t}", file=stream)

        print("\t_prop_map_put_ = {", file=stream)
        # These are "Invoke" args
        names = list(self.propMap.keys())
        names.sort()
        for key in names:
            entry = self.propMap[key]
            if generator.bBuildHidden or not entry.hidden:
                lkey = key.lower()
                details = entry.desc
                # If default arg is None, write an empty tuple
                defArgDesc = build.MakeDefaultArgRepr(details[2])
                if defArgDesc is None:
                    defArgDesc = ""
                else:
                    defArgDesc = defArgDesc + ","
                print(
                    '\t\t"%s" : ((%s, LCID, %d, 0),(%s)),'
                    % (
                        build.MakePublicAttributeName(key),
                        details[0],
                        pythoncom.DISPATCH_PROPERTYPUT,
                        defArgDesc,
                    ),
                    file=stream,
                )

        names = list(self.propMapPut.keys())
        names.sort()
        for key in names:
            entry = self.propMapPut[key]
            if generator.bBuildHidden or not entry.hidden:
                details = entry.desc
                defArgDesc = MakeDefaultArgsForPropertyPut(details[2])
                print(
                    '\t\t"%s": ((%s, LCID, %d, 0),%s),'
                    % (
                        build.MakePublicAttributeName(key),
                        details[0],
                        details[4],
                        defArgDesc,
                    ),
                    file=stream,
                )
        print("\t}", file=stream)

        if specialItems["value"]:
            entry, invoketype, propArgs = specialItems["value"]
            if propArgs is None:
                typename = "method"
                ret = self.MakeFuncMethod(entry, "__call__")
            else:
                typename = "property"
                ret = [
                    "\tdef __call__(self):\n\t\treturn self._ApplyTypes_(*%s)"
                    % propArgs
                ]
            print(
                "\t# Default %s for this class is '%s'" % (typename, entry.names[0]),
                file=stream,
            )
            for line in ret:
                print(line, file=stream)
            print("\tdef __str__(self, *args):", file=stream)
            print("\t\treturn str(self.__call__(*args))", file=stream)
            print("\tdef __int__(self, *args):", file=stream)
            print("\t\treturn int(self.__call__(*args))", file=stream)

        # _NewEnum (DISPID_NEWENUM) does not appear in typelib for many office objects,
        # but it can still be retrieved at runtime, so  always create __iter__.
        # Also, some of those same objects use 1-based indexing, causing the old-style
        # __getitem__ iteration to fail for index 0 where the dynamic iteration succeeds.
        if specialItems["_newenum"]:
            enumEntry, invoketype, propArgs = specialItems["_newenum"]
            assert enumEntry.desc.desckind == pythoncom.DESCKIND_FUNCDESC
            invkind = enumEntry.desc.invkind
            # ??? Wouldn't this be the resultCLSID for the iterator itself, rather than the resultCLSID
            #  for the result of each Next() call, which is what it's used for ???
            resultCLSID = enumEntry.GetResultCLSIDStr()
        else:
            invkind = pythoncom.DISPATCH_METHOD | pythoncom.DISPATCH_PROPERTYGET
            resultCLSID = "None"
        # If we dont have a good CLSID for the enum result, assume it is the same as the Item() method.
        if resultCLSID == "None" and "Item" in self.mapFuncs:
            resultCLSID = self.mapFuncs["Item"].GetResultCLSIDStr()
        print("\tdef __iter__(self):", file=stream)
        print('\t\t"Return a Python iterator for this object"', file=stream)
        print("\t\ttry:", file=stream)
        print(
            "\t\t\tob = self._oleobj_.InvokeTypes(%d,LCID,%d,(13, 10),())"
            % (pythoncom.DISPID_NEWENUM, invkind),
            file=stream,
        )
        print("\t\texcept pythoncom.error:", file=stream)
        print(
            '\t\t\traise TypeError("This object does not support enumeration")',
            file=stream,
        )
        # Iterator is wrapped as PyIEnumVariant, and each result of __next__ is Dispatch'ed if necessary
        print(
            "\t\treturn win32com.client.util.Iterator(ob, %s)" % resultCLSID,
            file=stream,
        )

        if specialItems["item"]:
            entry, invoketype, propArgs = specialItems["item"]
            resultCLSID = entry.GetResultCLSIDStr()
            print(
                "\t#This class has Item property/method which allows indexed access with the object[key] syntax.",
                file=stream,
            )
            print(
                "\t#Some objects will accept a string or other type of key in addition to integers.",
                file=stream,
            )
            print(
                "\t#Note that many Office objects do not use zero-based indexing.",
                file=stream,
            )
            print("\tdef __getitem__(self, key):", file=stream)
            print(
                '\t\treturn self._get_good_object_(self._oleobj_.Invoke(*(%d, LCID, %d, 1, key)), "Item", %s)'
                % (entry.desc.memid, invoketype, resultCLSID),
                file=stream,
            )

        if specialItems["count"]:
            entry, invoketype, propArgs = specialItems["count"]
            if propArgs is None:
                typename = "method"
                ret = self.MakeFuncMethod(entry, "__len__")
            else:
                typename = "property"
                ret = [
                    "\tdef __len__(self):\n\t\treturn self._ApplyTypes_(*%s)" % propArgs
                ]
            print(
                "\t#This class has Count() %s - allow len(ob) to provide this"
                % (typename),
                file=stream,
            )
            for line in ret:
                print(line, file=stream)
            # Also include a __nonzero__
            print(
                "\t#This class has a __len__ - this is needed so 'if object:' always returns TRUE.",
                file=stream,
            )
            print("\tdef __nonzero__(self):", file=stream)
            print("\t\treturn True", file=stream)


class CoClassItem(build.OleItem, WritableItem):
    order = 5
    typename = "COCLASS"

    def __init__(self, typeinfo, attr, doc=None, sources=[], interfaces=[], bForUser=1):
        build.OleItem.__init__(self, doc)
        self.clsid = attr[0]
        self.sources = sources
        self.interfaces = interfaces
        self.bIsDispatch = 1  # Pretend it is so it is written to the class map.

    def WriteClass(self, generator):
        generator.checkWriteCoClassBaseClass()
        doc = self.doc
        stream = generator.file
        if generator.generate_type == GEN_DEMAND_CHILD:
            # Some special imports we must setup.
            referenced_items = []
            for ref, flag in self.sources:
                referenced_items.append(ref)
            for ref, flag in self.interfaces:
                referenced_items.append(ref)
            print("import sys", file=stream)
            for ref in referenced_items:
                print(
                    "__import__('%s.%s')" % (generator.base_mod_name, ref.python_name),
                    file=stream,
                )
                print(
                    "%s = sys.modules['%s.%s'].%s"
                    % (
                        ref.python_name,
                        generator.base_mod_name,
                        ref.python_name,
                        ref.python_name,
                    ),
                    file=stream,
                )
                # And pretend we have written it - the name is now available as if we had!
                ref.bWritten = 1
        try:
            progId = pythoncom.ProgIDFromCLSID(self.clsid)
            print("# This CoClass is known by the name '%s'" % (progId), file=stream)
        except pythoncom.com_error:
            pass
        print(
            "class %s(CoClassBaseClass): # A CoClass" % (self.python_name), file=stream
        )
        if doc and doc[1]:
            print("\t# " + doc[1], file=stream)
        print("\tCLSID = %r" % (self.clsid,), file=stream)
        print("\tcoclass_sources = [", file=stream)
        defItem = None
        for item, flag in self.sources:
            if flag & pythoncom.IMPLTYPEFLAG_FDEFAULT:
                defItem = item
            # If we have written a Python class, reference the name -
            # otherwise just the IID.
            if item.bWritten:
                key = item.python_name
            else:
                key = repr(str(item.clsid))  # really the iid.
            print("\t\t%s," % (key), file=stream)
        print("\t]", file=stream)
        if defItem:
            if defItem.bWritten:
                defName = defItem.python_name
            else:
                defName = repr(str(defItem.clsid))  # really the iid.
            print("\tdefault_source = %s" % (defName,), file=stream)
        print("\tcoclass_interfaces = [", file=stream)
        defItem = None
        for item, flag in self.interfaces:
            if flag & pythoncom.IMPLTYPEFLAG_FDEFAULT:  # and dual:
                defItem = item
            # If we have written a class, reference its name, otherwise the IID
            if item.bWritten:
                key = item.python_name
            else:
                key = repr(str(item.clsid))  # really the iid.
            print("\t\t%s," % (key,), file=stream)
        print("\t]", file=stream)
        if defItem:
            if defItem.bWritten:
                defName = defItem.python_name
            else:
                defName = repr(str(defItem.clsid))  # really the iid.
            print("\tdefault_interface = %s" % (defName,), file=stream)
        self.bWritten = 1
        print(file=stream)


class GeneratorProgress:
    def __init__(self):
        pass

    def Starting(self, tlb_desc):
        """Called when the process starts."""
        self.tlb_desc = tlb_desc

    def Finished(self):
        """Called when the process is complete."""

    def SetDescription(self, desc, maxticks=None):
        """We are entering a major step.  If maxticks, then this
        is how many ticks we expect to make until finished
        """

    def Tick(self, desc=None):
        """Minor progress step.  Can provide new description if necessary"""

    def VerboseProgress(self, desc):
        """Verbose/Debugging output."""

    def LogWarning(self, desc):
        """If a warning is generated"""

    def LogBeginGenerate(self, filename):
        pass

    def Close(self):
        pass


class Generator:
    def __init__(
        self,
        typelib,
        sourceFilename,
        progressObject,
        bBuildHidden=1,
        bUnicodeToString=None,
    ):
        assert bUnicodeToString is None, "this is deprecated and will go away"
        self.bHaveWrittenDispatchBaseClass = 0
        self.bHaveWrittenCoClassBaseClass = 0
        self.bHaveWrittenEventBaseClass = 0
        self.typelib = typelib
        self.sourceFilename = sourceFilename
        self.bBuildHidden = bBuildHidden
        self.progress = progressObject
        # These 2 are later additions and most of the code still 'print's...
        self.file = None

    def CollectOleItemInfosFromType(self):
        ret = []
        for i in range(self.typelib.GetTypeInfoCount()):
            info = self.typelib.GetTypeInfo(i)
            infotype = self.typelib.GetTypeInfoType(i)
            doc = self.typelib.GetDocumentation(i)
            attr = info.GetTypeAttr()
            ret.append((info, infotype, doc, attr))
        return ret

    def _Build_CoClass(self, type_info_tuple):
        info, infotype, doc, attr = type_info_tuple
        # find the source and dispinterfaces for the coclass
        child_infos = []
        for j in range(attr[8]):
            flags = info.GetImplTypeFlags(j)
            try:
                refType = info.GetRefTypeInfo(info.GetRefTypeOfImplType(j))
            except pythoncom.com_error:
                # Can't load a dependent typelib?
                continue
            refAttr = refType.GetTypeAttr()
            child_infos.append(
                (
                    info,
                    refAttr.typekind,
                    refType,
                    refType.GetDocumentation(-1),
                    refAttr,
                    flags,
                )
            )

        # Done generating children - now the CoClass itself.
        newItem = CoClassItem(info, attr, doc)
        return newItem, child_infos

    def _Build_CoClassChildren(self, coclass, coclass_info, oleItems, vtableItems):
        sources = {}
        interfaces = {}
        for info, info_type, refType, doc, refAttr, flags in coclass_info:
            #          sys.stderr.write("Attr typeflags for coclass referenced object %s=%d (%d), typekind=%d\n" % (name, refAttr.wTypeFlags, refAttr.wTypeFlags & pythoncom.TYPEFLAG_FDUAL,refAttr.typekind))
            if refAttr.typekind == pythoncom.TKIND_DISPATCH or (
                refAttr.typekind == pythoncom.TKIND_INTERFACE
                and refAttr[11] & pythoncom.TYPEFLAG_FDISPATCHABLE
            ):
                clsid = refAttr[0]
                if clsid in oleItems:
                    dispItem = oleItems[clsid]
                else:
                    dispItem = DispatchItem(refType, refAttr, doc)
                    oleItems[dispItem.clsid] = dispItem
                dispItem.coclass_clsid = coclass.clsid
                if flags & pythoncom.IMPLTYPEFLAG_FSOURCE:
                    dispItem.bIsSink = 1
                    sources[dispItem.clsid] = (dispItem, flags)
                else:
                    interfaces[dispItem.clsid] = (dispItem, flags)
                # If dual interface, make do that too.
                if clsid not in vtableItems and refAttr[11] & pythoncom.TYPEFLAG_FDUAL:
                    refType = refType.GetRefTypeInfo(refType.GetRefTypeOfImplType(-1))
                    refAttr = refType.GetTypeAttr()
                    assert (
                        refAttr.typekind == pythoncom.TKIND_INTERFACE
                    ), "must be interface bynow!"
                    vtableItem = VTableItem(refType, refAttr, doc)
                    vtableItems[clsid] = vtableItem
        coclass.sources = list(sources.values())
        coclass.interfaces = list(interfaces.values())

    def _Build_Interface(self, type_info_tuple):
        info, infotype, doc, attr = type_info_tuple
        oleItem = vtableItem = None
        if infotype == pythoncom.TKIND_DISPATCH or (
            infotype == pythoncom.TKIND_INTERFACE
            and attr[11] & pythoncom.TYPEFLAG_FDISPATCHABLE
        ):
            oleItem = DispatchItem(info, attr, doc)
            # If this DISPATCH interface dual, then build that too.
            if attr.wTypeFlags & pythoncom.TYPEFLAG_FDUAL:
                # Get the vtable interface
                refhtype = info.GetRefTypeOfImplType(-1)
                info = info.GetRefTypeInfo(refhtype)
                attr = info.GetTypeAttr()
                infotype = pythoncom.TKIND_INTERFACE
            else:
                infotype = None
        assert infotype in [
            None,
            pythoncom.TKIND_INTERFACE,
        ], "Must be a real interface at this point"
        if infotype == pythoncom.TKIND_INTERFACE:
            vtableItem = VTableItem(info, attr, doc)
        return oleItem, vtableItem

    def BuildOleItemsFromType(self):
        assert (
            self.bBuildHidden
        ), "This code doesnt look at the hidden flag - I thought everyone set it true!?!?!"
        oleItems = {}
        enumItems = {}
        recordItems = {}
        vtableItems = {}

        for type_info_tuple in self.CollectOleItemInfosFromType():
            info, infotype, doc, attr = type_info_tuple
            clsid = attr[0]
            if infotype == pythoncom.TKIND_ENUM or infotype == pythoncom.TKIND_MODULE:
                newItem = EnumerationItem(info, attr, doc)
                enumItems[newItem.doc[0]] = newItem
            # We never hide interfaces (MSAccess, for example, nominates interfaces as
            # hidden, assuming that you only ever use them via the CoClass)
            elif infotype in [pythoncom.TKIND_DISPATCH, pythoncom.TKIND_INTERFACE]:
                if clsid not in oleItems:
                    oleItem, vtableItem = self._Build_Interface(type_info_tuple)
                    oleItems[clsid] = oleItem  # Even "None" goes in here.
                    if vtableItem is not None:
                        vtableItems[clsid] = vtableItem
            elif (
                infotype == pythoncom.TKIND_RECORD or infotype == pythoncom.TKIND_UNION
            ):
                newItem = RecordItem(info, attr, doc)
                recordItems[newItem.clsid] = newItem
            elif infotype == pythoncom.TKIND_ALIAS:
                # We dont care about alias' - handled intrinsicly.
                continue
            elif infotype == pythoncom.TKIND_COCLASS:
                newItem, child_infos = self._Build_CoClass(type_info_tuple)
                self._Build_CoClassChildren(newItem, child_infos, oleItems, vtableItems)
                oleItems[newItem.clsid] = newItem
            else:
                self.progress.LogWarning("Unknown TKIND found: %d" % infotype)

        return oleItems, enumItems, recordItems, vtableItems

    def open_writer(self, filename, encoding="mbcs"):
        # A place to put code to open a file with the appropriate encoding.
        # Does *not* set self.file - just opens and returns a file.
        # Actually returns a handle to a temp file - finish_writer then deletes
        # the filename asked for and puts everything back in place.  This
        # is so errors don't leave a 1/2 generated file around causing bizarre
        # errors later, and so that multiple processes writing the same file
        # don't step on each others' toes.
        # Could be a classmethod one day...
        temp_filename = self.get_temp_filename(filename)
        return open(temp_filename, "wt", encoding=encoding)

    def finish_writer(self, filename, f, worked):
        f.close()
        try:
            os.unlink(filename)
        except os.error:
            pass
        temp_filename = self.get_temp_filename(filename)
        if worked:
            try:
                os.rename(temp_filename, filename)
            except os.error:
                # If we are really unlucky, another process may have written the
                # file in between our calls to os.unlink and os.rename. So try
                # again, but only once.
                # There are still some race conditions, but they seem difficult to
                # fix, and they probably occur much less frequently:
                # * The os.rename failure could occur more than once if more than
                #   two processes are involved.
                # * In between os.unlink and os.rename, another process could try
                #   to import the module, having seen that it already exists.
                # * If another process starts a COM server while we are still
                #   generating __init__.py, that process sees that the folder
                #   already exists and assumes that __init__.py is already there
                #   as well.
                try:
                    os.unlink(filename)
                except os.error:
                    pass
                os.rename(temp_filename, filename)
        else:
            os.unlink(temp_filename)

    def get_temp_filename(self, filename):
        return "%s.%d.temp" % (filename, os.getpid())

    def generate(self, file, is_for_demand=0):
        if is_for_demand:
            self.generate_type = GEN_DEMAND_BASE
        else:
            self.generate_type = GEN_FULL
        self.file = file
        self.do_generate()
        self.file = None
        self.progress.Finished()

    def do_gen_file_header(self):
        la = self.typelib.GetLibAttr()
        moduleDoc = self.typelib.GetDocumentation(-1)
        docDesc = ""
        if moduleDoc[1]:
            docDesc = moduleDoc[1]

        # Reset all the 'per file' state
        self.bHaveWrittenDispatchBaseClass = 0
        self.bHaveWrittenCoClassBaseClass = 0
        self.bHaveWrittenEventBaseClass = 0
        # You must provide a file correctly configured for writing unicode.
        # We assert this is it may indicate somewhere in pywin32 that needs
        # upgrading.
        assert self.file.encoding, self.file
        encoding = self.file.encoding  # or "mbcs"

        print("# -*- coding: %s -*-" % (encoding,), file=self.file)
        print("# Created by makepy.py version %s" % (makepy_version,), file=self.file)
        print(
            "# By python version %s" % (sys.version.replace("\n", "-"),), file=self.file
        )
        if self.sourceFilename:
            print(
                "# From type library '%s'" % (os.path.split(self.sourceFilename)[1],),
                file=self.file,
            )
        print("# On %s" % time.ctime(time.time()), file=self.file)

        print(build._makeDocString(docDesc), file=self.file)

        print("makepy_version =", repr(makepy_version), file=self.file)
        print("python_version = 0x%x" % (sys.hexversion,), file=self.file)
        print(file=self.file)
        print(
            "import win32com.client.CLSIDToClass, pythoncom, pywintypes", file=self.file
        )
        print("import win32com.client.util", file=self.file)
        print("from pywintypes import IID", file=self.file)
        print("from win32com.client import Dispatch", file=self.file)
        print(file=self.file)
        print(
            "# The following 3 lines may need tweaking for the particular server",
            file=self.file,
        )
        print(
            "# Candidates are pythoncom.Missing, .Empty and .ArgNotFound",
            file=self.file,
        )
        print("defaultNamedOptArg=pythoncom.Empty", file=self.file)
        print("defaultNamedNotOptArg=pythoncom.Empty", file=self.file)
        print("defaultUnnamedArg=pythoncom.Empty", file=self.file)
        print(file=self.file)
        print("CLSID = " + repr(la[0]), file=self.file)
        print("MajorVersion = " + str(la[3]), file=self.file)
        print("MinorVersion = " + str(la[4]), file=self.file)
        print("LibraryFlags = " + str(la[5]), file=self.file)
        print("LCID = " + hex(la[1]), file=self.file)
        print(file=self.file)

    def do_generate(self):
        moduleDoc = self.typelib.GetDocumentation(-1)
        stream = self.file
        docDesc = ""
        if moduleDoc[1]:
            docDesc = moduleDoc[1]
        self.progress.Starting(docDesc)
        self.progress.SetDescription("Building definitions from type library...")

        self.do_gen_file_header()

        oleItems, enumItems, recordItems, vtableItems = self.BuildOleItemsFromType()

        self.progress.SetDescription(
            "Generating...", len(oleItems) + len(enumItems) + len(vtableItems)
        )

        # Generate the constants and their support.
        if enumItems:
            print("class constants:", file=stream)
            items = list(enumItems.values())
            items.sort()
            num_written = 0
            for oleitem in items:
                num_written += oleitem.WriteEnumerationItems(stream)
                self.progress.Tick()
            if not num_written:
                print("\tpass", file=stream)
            print(file=stream)

        if self.generate_type == GEN_FULL:
            items = [l for l in oleItems.values() if l is not None]
            items.sort()
            for oleitem in items:
                self.progress.Tick()
                oleitem.WriteClass(self)

            items = list(vtableItems.values())
            items.sort()
            for oleitem in items:
                self.progress.Tick()
                oleitem.WriteClass(self)
        else:
            self.progress.Tick(len(oleItems) + len(vtableItems))

        print("RecordMap = {", file=stream)
        for record in recordItems.values():
            if record.clsid == pythoncom.IID_NULL:
                print(
                    "\t###%s: %s, # Record disabled because it doesn't have a non-null GUID"
                    % (repr(record.doc[0]), repr(str(record.clsid))),
                    file=stream,
                )
            else:
                print(
                    "\t%s: %s," % (repr(record.doc[0]), repr(str(record.clsid))),
                    file=stream,
                )
        print("}", file=stream)
        print(file=stream)

        # Write out _all_ my generated CLSID's in the map
        if self.generate_type == GEN_FULL:
            print("CLSIDToClassMap = {", file=stream)
            for item in oleItems.values():
                if item is not None and item.bWritten:
                    print(
                        "\t'%s' : %s," % (str(item.clsid), item.python_name),
                        file=stream,
                    )
            print("}", file=stream)
            print("CLSIDToPackageMap = {}", file=stream)
            print(
                "win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )",
                file=stream,
            )
            print("VTablesToPackageMap = {}", file=stream)
            print("VTablesToClassMap = {", file=stream)
            for item in vtableItems.values():
                print("\t'%s' : '%s'," % (item.clsid, item.python_name), file=stream)
            print("}", file=stream)
            print(file=stream)

        else:
            print("CLSIDToClassMap = {}", file=stream)
            print("CLSIDToPackageMap = {", file=stream)
            for item in oleItems.values():
                if item is not None:
                    print(
                        "\t'%s' : %s," % (str(item.clsid), repr(item.python_name)),
                        file=stream,
                    )
            print("}", file=stream)
            print("VTablesToClassMap = {}", file=stream)
            print("VTablesToPackageMap = {", file=stream)
            for item in vtableItems.values():
                print("\t'%s' : '%s'," % (item.clsid, item.python_name), file=stream)
            print("}", file=stream)
            print(file=stream)

        print(file=stream)
        # Bit of a hack - build a temp map of iteItems + vtableItems - coClasses
        map = {}
        for item in oleItems.values():
            if item is not None and not isinstance(item, CoClassItem):
                map[item.python_name] = item.clsid
        for item in vtableItems.values():  # No nones or CoClasses in this map
            map[item.python_name] = item.clsid

        print("NamesToIIDMap = {", file=stream)
        for name, iid in map.items():
            print("\t'%s' : '%s'," % (name, iid), file=stream)
        print("}", file=stream)
        print(file=stream)

        if enumItems:
            print(
                "win32com.client.constants.__dicts__.append(constants.__dict__)",
                file=stream,
            )
        print(file=stream)

    def generate_child(self, child, dir):
        "Generate a single child.  May force a few children to be built as we generate deps"
        self.generate_type = GEN_DEMAND_CHILD

        la = self.typelib.GetLibAttr()
        lcid = la[1]
        clsid = la[0]
        major = la[3]
        minor = la[4]
        self.base_mod_name = (
            "win32com.gen_py." + str(clsid)[1:-1] + "x%sx%sx%s" % (lcid, major, minor)
        )
        try:
            # Process the type library's CoClass objects, looking for the
            # specified name, or where a child has the specified name.
            # This ensures that all interesting things (including event interfaces)
            # are generated correctly.
            oleItems = {}
            vtableItems = {}
            infos = self.CollectOleItemInfosFromType()
            found = 0
            for type_info_tuple in infos:
                info, infotype, doc, attr = type_info_tuple
                if infotype == pythoncom.TKIND_COCLASS:
                    coClassItem, child_infos = self._Build_CoClass(type_info_tuple)
                    found = build.MakePublicAttributeName(doc[0]) == child
                    if not found:
                        # OK, check the child interfaces
                        for (
                            info,
                            info_type,
                            refType,
                            doc,
                            refAttr,
                            flags,
                        ) in child_infos:
                            if build.MakePublicAttributeName(doc[0]) == child:
                                found = 1
                                break
                    if found:
                        oleItems[coClassItem.clsid] = coClassItem
                        self._Build_CoClassChildren(
                            coClassItem, child_infos, oleItems, vtableItems
                        )
                        break
            if not found:
                # Doesn't appear in a class defn - look in the interface objects for it
                for type_info_tuple in infos:
                    info, infotype, doc, attr = type_info_tuple
                    if infotype in [
                        pythoncom.TKIND_INTERFACE,
                        pythoncom.TKIND_DISPATCH,
                    ]:
                        if build.MakePublicAttributeName(doc[0]) == child:
                            found = 1
                            oleItem, vtableItem = self._Build_Interface(type_info_tuple)
                            oleItems[clsid] = oleItem  # Even "None" goes in here.
                            if vtableItem is not None:
                                vtableItems[clsid] = vtableItem

            assert (
                found
            ), "Cant find the '%s' interface in the CoClasses, or the interfaces" % (
                child,
            )
            # Make a map of iid: dispitem, vtableitem)
            items = {}
            for key, value in oleItems.items():
                items[key] = (value, None)
            for key, value in vtableItems.items():
                existing = items.get(key, None)
                if existing is not None:
                    new_val = existing[0], value
                else:
                    new_val = None, value
                items[key] = new_val

            self.progress.SetDescription("Generating...", len(items))
            for oleitem, vtableitem in items.values():
                an_item = oleitem or vtableitem
                assert not self.file, "already have a file?"
                # like makepy.py, we gen to a .temp file so failure doesn't
                # leave a 1/2 generated mess.
                out_name = os.path.join(dir, an_item.python_name) + ".py"
                worked = False
                self.file = self.open_writer(out_name)
                try:
                    if oleitem is not None:
                        self.do_gen_child_item(oleitem)
                    if vtableitem is not None:
                        self.do_gen_child_item(vtableitem)
                    self.progress.Tick()
                    worked = True
                finally:
                    self.finish_writer(out_name, self.file, worked)
                    self.file = None
        finally:
            self.progress.Finished()

    def do_gen_child_item(self, oleitem):
        moduleDoc = self.typelib.GetDocumentation(-1)
        docDesc = ""
        if moduleDoc[1]:
            docDesc = moduleDoc[1]
        self.progress.Starting(docDesc)
        self.progress.SetDescription("Building definitions from type library...")
        self.do_gen_file_header()
        oleitem.WriteClass(self)
        if oleitem.bWritten:
            print(
                'win32com.client.CLSIDToClass.RegisterCLSID( "%s", %s )'
                % (oleitem.clsid, oleitem.python_name),
                file=self.file,
            )

    def checkWriteDispatchBaseClass(self):
        if not self.bHaveWrittenDispatchBaseClass:
            print("from win32com.client import DispatchBaseClass", file=self.file)
            self.bHaveWrittenDispatchBaseClass = 1

    def checkWriteCoClassBaseClass(self):
        if not self.bHaveWrittenCoClassBaseClass:
            print("from win32com.client import CoClassBaseClass", file=self.file)
            self.bHaveWrittenCoClassBaseClass = 1

    def checkWriteEventBaseClass(self):
        # Not a base class as such...
        if not self.bHaveWrittenEventBaseClass:
            # Nothing to do any more!
            self.bHaveWrittenEventBaseClass = 1


if __name__ == "__main__":
    print("This is a worker module.  Please use makepy to generate Python files.")



# win2kras used to be an extension module with wrapped the "new" RAS functions
# in Windows 2000, so win32ras could still be used on NT/etc.
# I think in 2021 we can be confident pywin32 is not used on earlier OSs, so
# that functionality is now in win32ras.
#
# This exists just to avoid breaking old scripts.
from win32ras import *
